@@ -42,7 +42,7 @@ struct _GIOPRecvBuffer {
 
 	GIOPVersion giop_version;
 	gulong      left_to_read;
-	gboolean    free_body : 1;
+	guchar      free_body : 1;
 };
 
 #define         giop_msg_conversion_needed(msg)     giop_endian_conversion_needed(GIOP_MSG(msg)->header.flags)
@@ -71,11 +71,11 @@ CORBA_unsigned_long         giop_recv_buffer_get_request_id (GIOPRecvBuffer *buf
 char                       *giop_recv_buffer_get_opname     (GIOPRecvBuffer *buf);
 CORBA_sequence_CORBA_octet *giop_recv_buffer_get_objkey     (GIOPRecvBuffer *buf);
 void                        giop_recv_list_zap              (GIOPConnection *cnx);
-gboolean                    giop_connection_handle_input    (LINCConnection *lcnx);
+gboolean                    giop_connection_handle_input    (LinkConnection *lcnx);
 void                        giop_connection_destroy_frags   (GIOPConnection *cnx);
 
 gpointer                    giop_recv_thread_fn             (gpointer        data);
-gboolean                    giop_recv_handle_queued_input   (void);
+void                        giop_recv_handle_queued_input   (void);
 
 #endif /* ORBIT2_INTERNAL_API */
 
