@@ -374,42 +374,54 @@ orbit_idl_print_node(IDL_tree node, int indent_level)
 static void
 IDL_tree_traverse_helper(IDL_tree p, GFunc f,
 			 gconstpointer func_data,
-			 GHashTable *visited_nodes)
+			 GHashTable *visited_nodes,
+			 gboolean    include_self)
 {
 	IDL_tree curitem;
 
-	if(g_hash_table_lookup(visited_nodes, p))
+	if (g_hash_table_lookup (visited_nodes, p))
 		return;
 
-	g_hash_table_insert(visited_nodes, p, ((gpointer)1));
+	g_hash_table_insert (visited_nodes, p, ((gpointer)1));
 
-	for(curitem = IDL_INTERFACE(p).inheritance_spec; curitem;
-	    curitem = IDL_LIST(curitem).next) {
-		IDL_tree_traverse_helper(IDL_get_parent_node(IDL_LIST(curitem).data, IDLN_INTERFACE, NULL), f, func_data, visited_nodes);
+	for (curitem = IDL_INTERFACE (p).inheritance_spec; curitem;
+	     curitem = IDL_LIST (curitem).next) {
+		IDL_tree_traverse_helper (IDL_get_parent_node 
+			(IDL_LIST (curitem).data, IDLN_INTERFACE, NULL), f, func_data, visited_nodes, TRUE);
 	}
 
-	f(p, (gpointer)func_data);
+	if (include_self)
+		f(p, (gpointer)func_data);
 }
 
 void
-IDL_tree_traverse_parents(IDL_tree p,
-			  GFunc f,
-			  gconstpointer func_data)
+IDL_tree_traverse_parents_full (IDL_tree      p,
+				GFunc         f,
+				gconstpointer func_data,
+				gboolean      include_self)
 {
-	GHashTable *visited_nodes = g_hash_table_new(NULL, g_direct_equal);
+	GHashTable *visited_nodes = g_hash_table_new (NULL, g_direct_equal);
 
-	if(!(p && f))
+	if (!(p && f))
 		return;
 
-	if(IDL_NODE_TYPE(p) != IDLN_INTERFACE)
-		p = IDL_get_parent_node(p, IDLN_INTERFACE, NULL);
+	if (IDL_NODE_TYPE(p) != IDLN_INTERFACE)
+		p = IDL_get_parent_node (p, IDLN_INTERFACE, NULL);
 
-	if(!p)
+	if (!p)
 		return;
 
-	IDL_tree_traverse_helper(p, f, func_data, visited_nodes);
+	IDL_tree_traverse_helper (p, f, func_data, visited_nodes, include_self);
 
-	g_hash_table_destroy(visited_nodes);
+	g_hash_table_destroy (visited_nodes);
+}
+
+void
+IDL_tree_traverse_parents (IDL_tree p,
+			   GFunc f,
+			   gconstpointer func_data)
+{
+	IDL_tree_traverse_parents_full (p, f, func_data, TRUE);
 }
 
 /* For use by below function */
