@@ -186,6 +186,12 @@ giop_connection_initiate (gpointer orb_data,
 			 NULL);
 }
 
+LinkConnectionStatus
+giop_connection_try_reconnect (GIOPConnection *cnx)
+{
+	return link_connection_try_reconnect (LINK_CONNECTION (cnx));
+}
+
 void
 giop_connections_shutdown (void)
 {
@@ -216,3 +222,4 @@ giop_connections_shutdown (void)
 		g_warning ("Wierd; new connections opened while shutting down");
 #endif
 }
+
