@@ -1,6 +1,7 @@
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
+#include <stdarg.h>
 
 #include <orbit/orbit.h>
 #include <orb-core-export.h>
@@ -11,9 +12,16 @@
 #include "../GIOP/giop-debug.h"
 #include "../GIOP/giop-recv-buffer.h"
 #include "poa-macros.h"
-
+#include "poa-private.h"
 #include "orbit-poa.h"
 
+#define POA_LOCK(poa)   LINK_MUTEX_LOCK(poa->base.lock)
+#define POA_UNLOCK(poa) LINK_MUTEX_UNLOCK(poa->base.lock)
+
+static GMutex     *ORBit_class_assignment_lock    = NULL;
+static GHashTable *ORBit_class_assignments        = NULL;
+static guint       ORBit_class_assignment_counter = 0;
+
 static PortableServer_Servant ORBit_POA_ServantManager_use_servant(
 				     PortableServer_POA poa,
 				     ORBit_POAObject pobj,
@@ -55,10 +63,7 @@ static void               ORBit_POA_deactivate_object       (PortableServer_POA
 							     ORBit_POAObject     pobj,
 							     CORBA_boolean       do_etherealize,
 							     CORBA_boolean       is_cleanup);
-
-static GHashTable *ORBit_class_assignments = NULL;
-static guint ORBit_class_assignment_counter = 0;
-
+     
 /* PortableServer_Current interface */
 static void
 ORBit_POACurrent_free_fn (ORBit_RootObject obj_in)
@@ -95,24 +100,36 @@ static ORBit_POAObject
 ORBit_POACurrent_get_object (PortableServer_Current  obj,
 			     CORBA_Environment      *ev)
 {
+	ORBit_POAObject pobj = CORBA_OBJECT_NIL;
+
 	g_assert (obj && obj->parent.interface->type == ORBIT_ROT_POACURRENT);
 
-	poa_exception_val_if_fail (obj->orb->current_invocations != NULL,
-				   ex_PortableServer_Current_NoContext,
-				   CORBA_OBJECT_NIL);
+	LINK_MUTEX_LOCK (obj->orb->lock);
+	if (obj->orb->current_invocations == NULL)
+		CORBA_exception_set_system
+			(ev, ex_PortableServer_Current_NoContext,
+			 CORBA_COMPLETED_NO);
+	else
+		pobj = (ORBit_POAObject) obj->orb->current_invocations->data;
+	LINK_MUTEX_UNLOCK (obj->orb->lock);
 
-	return (ORBit_POAObject) obj->orb->current_invocations->data;
+	return pobj;
 }
 
 PortableServer_ClassInfo *
 ORBit_classinfo_lookup (const char *type_id)
 {
-	if (!ORBit_class_assignments)
-		return NULL;
+	PortableServer_ClassInfo *ci = NULL;
 
-	return g_hash_table_lookup (ORBit_class_assignments, type_id);
+	LINK_MUTEX_LOCK   (ORBit_class_assignment_lock);
+	if (ORBit_class_assignments)
+		ci = g_hash_table_lookup (ORBit_class_assignments, type_id);
+	LINK_MUTEX_UNLOCK (ORBit_class_assignment_lock);
+
+	return ci;
 }
 
+/* Deprecated & not thread-safe */
 void
 ORBit_classinfo_register (PortableServer_ClassInfo *ci)
 {
@@ -133,28 +150,79 @@ ORBit_classinfo_register (PortableServer_ClassInfo *ci)
 			     (gpointer) ci->class_name, ci);
 }
 
+void
+ORBit_skel_class_register (PortableServer_ClassInfo   *ci,
+			   PortableServer_ServantBase *servant,
+			   void                      (*opt_finalize) (PortableServer_Servant,
+								      CORBA_Environment *),
+			   CORBA_unsigned_long         class_offset,
+			   CORBA_unsigned_long         first_parent_id,
+			   ...)
+{
+	va_list args;
+	CORBA_unsigned_long id;
+
+	va_start (args, first_parent_id);
+
+	/*
+	 * FIXME: a double check/lock/write barrier pattern
+	 * would be far faster here.
+	 */
+	LINK_MUTEX_LOCK (ORBit_class_assignment_lock);
+
+	ORBit_classinfo_register (ci);
+	if (!ci->vepvmap) {
+		CORBA_unsigned_long offset;
+
+		/* FIXME: we could propagate the size nicely here */
+		ci->vepvmap = g_new0 (ORBit_VepvIdx, 
+				      *(ci->class_id) + 1);
+
+		ci->vepvmap[*(ci->class_id)] = class_offset;
+		
+		for (id = first_parent_id; id;) {
+
+			offset = va_arg (args, CORBA_unsigned_long);
+			
+			g_assert (id <= *(ci->class_id));
+			ci->vepvmap [id] = offset;
+
+			id = va_arg (args, CORBA_unsigned_long);
+		}
+	}
+	
+	LINK_MUTEX_UNLOCK (ORBit_class_assignment_lock);
+
+	if (!servant->vepv[0]->finalize)
+		servant->vepv[0]->finalize = opt_finalize;
+
+	ORBIT_SERVANT_SET_CLASSINFO (servant, ci);
+
+	va_end (args);
+}
+
 static void
-check_child_poa_inuse (char               *name,
-		       PortableServer_POA  poa,
-		       gboolean           *is_inuse)
+check_child_poa_inuse_T (char               *name,
+			 PortableServer_POA  poa,
+			 gboolean           *is_inuse)
 {
 	if (ORBit_POA_is_inuse (poa, CORBA_TRUE, NULL))
 		*is_inuse = TRUE;
 }
 
 static void
-check_object_inuse (PortableServer_ObjectId *oid,
-		    ORBit_POAObject          pobj, 
-		    gboolean                *is_inuse)
+check_object_inuse_T (PortableServer_ObjectId *oid,
+		      ORBit_POAObject          pobj, 
+		      gboolean                *is_inuse)
 {
 	if (pobj->use_cnt > 0)
 		*is_inuse = TRUE;
 }
 
-gboolean
-ORBit_POA_is_inuse (PortableServer_POA  poa,
-		    CORBA_boolean       consider_children,
-		    CORBA_Environment  *ev)
+static gboolean
+ORBit_POA_is_inuse_T (PortableServer_POA  poa,
+		      CORBA_boolean       consider_children,
+		      CORBA_Environment  *ev)
 {
 	gboolean is_inuse = FALSE;
 
@@ -163,19 +231,33 @@ ORBit_POA_is_inuse (PortableServer_POA  poa,
 
 	if (consider_children && poa->child_poas)
 		g_hash_table_foreach (poa->child_poas,
-				      (GHFunc) check_child_poa_inuse,
+				      (GHFunc) check_child_poa_inuse_T,
 				     &is_inuse);
 
 	if (!is_inuse && poa->oid_to_obj_map)
 		g_hash_table_foreach (poa->oid_to_obj_map,
-				      (GHFunc) check_object_inuse,
+				      (GHFunc) check_object_inuse_T,
 				      &is_inuse);
 
 	return is_inuse;
 }
 
-static PortableServer_ObjectId*
-ORBit_POA_new_system_objid (PortableServer_POA poa)
+gboolean
+ORBit_POA_is_inuse (PortableServer_POA  poa,
+		    CORBA_boolean       consider_children,
+		    CORBA_Environment  *ev)
+{
+	gboolean inuse;
+
+	POA_LOCK (poa);
+	inuse = ORBit_POA_is_inuse_T (poa, consider_children, ev);
+	POA_UNLOCK (poa);
+
+	return inuse;
+}
+
+static PortableServer_ObjectId *
+ORBit_POA_new_system_objid_T (PortableServer_POA poa)
 {
 	PortableServer_ObjectId *objid;
 
@@ -220,8 +302,8 @@ ORBit_POAObject_object_to_objkey (ORBit_POAObject pobj)
 }
 
 static ORBit_POAObject
-ORBit_POA_object_id_lookup (PortableServer_POA             poa,
-			    const PortableServer_ObjectId *oid)
+ORBit_POA_object_id_lookup_T (PortableServer_POA             poa,
+			      const PortableServer_ObjectId *oid)
 {
 	/* FIXME: needs some locking action */
 	return ORBit_RootObject_duplicate (
@@ -439,6 +521,8 @@ ORBit_POA_free_fn (ORBit_RootObject obj)
 	ORBit_ObjectAdaptor adaptor = (ORBit_ObjectAdaptor) obj;
 	PortableServer_POA  poa = (PortableServer_POA) obj;
 
+	giop_thread_key_release (obj);
+
 	if (adaptor->adaptor_key._buffer)
 		ORBit_free_T (adaptor->adaptor_key._buffer);
 
@@ -451,6 +535,9 @@ ORBit_POA_free_fn (ORBit_RootObject obj)
 	if (poa->name)
 		g_free (poa->name);
 
+	if (poa->base.lock)
+		g_mutex_free (poa->base.lock);
+
 	ORBit_RootObject_release_T (poa->orb);
 	ORBit_RootObject_release_T (poa->poa_manager);
 
@@ -534,6 +621,19 @@ ORBit_POA_set_policy (PortableServer_POA  poa,
 	}
 }
 
+static void
+ORBit_POA_copy_policies (const PortableServer_POA src,
+			 PortableServer_POA dest)
+{
+	dest->p_thread              = src->p_thread;
+	dest->p_lifespan            = src->p_lifespan;
+	dest->p_id_uniqueness       = src->p_id_uniqueness;
+	dest->p_id_assignment       = src->p_id_assignment;
+	dest->p_servant_retention   = src->p_servant_retention;
+	dest->p_request_processing  = src->p_request_processing;
+	dest->p_implicit_activation = src->p_implicit_activation;
+}
+
 static void
 ORBit_POA_set_policies (PortableServer_POA      poa,
 			const CORBA_PolicyList *policies,
@@ -608,11 +708,41 @@ ORBit_POA_new (CORBA_ORB                  orb,
 			(GHashFunc) ORBit_sequence_CORBA_octet_hash,
 			(GEqualFunc) ORBit_sequence_CORBA_octet_equal);
 
+	poa->base.lock = link_mutex_new ();
+
 	ORBit_POAManager_register_poa (manager, poa);
 
 	return ORBit_RootObject_duplicate (poa);
 }
 
+PortableServer_POA
+ORBit_POA_new_from (CORBA_ORB                  orb,
+		    PortableServer_POA         parent,
+		    const CORBA_char          *adaptor_name,
+		    const CORBA_PolicyList    *opt_policies,
+		    CORBA_Environment         *ev)
+{
+	PortableServer_POA poa;
+	
+	g_return_val_if_fail (parent != CORBA_OBJECT_NIL, CORBA_OBJECT_NIL);
+
+	poa = ORBit_POA_new (orb, adaptor_name, parent->poa_manager, NULL, ev);
+
+	g_return_val_if_fail (poa != CORBA_OBJECT_NIL, CORBA_OBJECT_NIL);
+
+	ORBit_POA_copy_policies (parent, poa);
+
+	if (opt_policies) {
+		int i;
+		for (i = 0; i < opt_policies->_length; i++)
+			ORBit_POA_set_policy (poa, opt_policies->_buffer[i]);
+	}
+
+	ORBit_POA_add_child (parent, poa);
+
+	return poa;
+}
+
 static CORBA_Object
 ORBit_POA_obj_to_ref (PortableServer_POA  poa,
 		      ORBit_POAObject     pobj,
@@ -634,11 +764,9 @@ ORBit_POA_obj_to_ref (PortableServer_POA  poa,
 	oid = pobj->object_id;
 
 	pobj->base.objref = ORBit_objref_new (poa->poa_manager->orb,
+					      &pobj->base,
 					      g_quark_from_string (type_id));
 
-	/* released by CORBA_Object_release */
-	pobj->base.objref->adaptor_obj = ORBit_RootObject_duplicate (pobj);
-
 	return ORBit_RootObject_duplicate (pobj->base.objref);
 }
 
@@ -681,6 +809,8 @@ ORBit_POAObject_release_cb (ORBit_RootObject robj)
 	/* object *must* be deactivated */
 	g_assert (pobj->servant == NULL);
 
+	giop_thread_key_release (robj);
+
 	object_id = pobj->object_id;
 	pobj->object_id = NULL;
 
@@ -723,9 +853,9 @@ ORBit_OAObject_Interface_type ORBit_POAObject_methods = {
  *  is allowed by the CORBA spec.
  */
 static ORBit_POAObject
-ORBit_POA_create_object (PortableServer_POA             poa,
-			 const PortableServer_ObjectId *objid,
-			 CORBA_Environment             *ev)
+ORBit_POA_create_object_T (PortableServer_POA             poa,
+			   const PortableServer_ObjectId *objid,
+			   CORBA_Environment             *ev)
 {
 	ORBit_POAObject newobj;
 
@@ -751,7 +881,7 @@ ORBit_POA_create_object (PortableServer_POA             poa,
 			memcpy (newobj->object_id->_buffer, objid->_buffer, objid->_length);
 		}
 		else
-			newobj->object_id = ORBit_POA_new_system_objid (poa);
+			newobj->object_id = ORBit_POA_new_system_objid_T (poa);
 	} else {
 		newobj->object_id           = PortableServer_ObjectId__alloc ();
 		newobj->object_id->_length  = objid->_length;
@@ -773,10 +903,10 @@ ORBit_POA_create_object (PortableServer_POA             poa,
  *  or NON_RETAIN.
  */
 static void
-ORBit_POA_activate_object (PortableServer_POA          poa, 
-			   ORBit_POAObject             pobj,
-			   PortableServer_ServantBase *servant, 
-			   CORBA_Environment          *ev) 
+ORBit_POA_activate_object_T (PortableServer_POA          poa, 
+			     ORBit_POAObject             pobj,
+			     PortableServer_ServantBase *servant, 
+			     CORBA_Environment          *ev) 
 {
 	PortableServer_ClassInfo *class_info;
 
@@ -785,7 +915,6 @@ ORBit_POA_activate_object (PortableServer_POA          poa,
 	g_assert (pobj->use_cnt == 0);
 
 	class_info = ORBIT_SERVANT_TO_CLASSINFO (servant);
-	g_assert (class_info->vepvmap);
 	pobj->vepvmap_cache = class_info->vepvmap;
 
 	pobj->servant = servant;
@@ -968,12 +1097,15 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 {
 	PortableServer_POA                   poa = pobj->poa;
 	PortableServer_ServantLocator_Cookie cookie;
-	PortableServer_ObjectId             *oid;
+	PortableServer_ObjectId             *oid = NULL;
 	PortableServer_ClassInfo            *klass;
 	ORBit_IMethod                       *m_data = NULL;
 	ORBitSmallSkeleton                   small_skel = NULL;
 	gpointer                             imp = NULL;
 
+	if (poa)
+		POA_LOCK (poa);
+
 	if (!poa || !poa->poa_manager)
 		CORBA_exception_set_system (
 			ev, ex_CORBA_OBJECT_NOT_EXIST, 
@@ -988,24 +1120,26 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 						
 				poa->held_requests = g_slist_prepend (
 					poa->held_requests, recv_buffer);
-			}
-			else
+				
+				POA_UNLOCK (poa);
+				return;
+			} else
 				CORBA_exception_set_system (
 					ev, ex_CORBA_TRANSIENT,
 					CORBA_COMPLETED_NO);
-			return;
+			break;
 			
 		case PortableServer_POAManager_DISCARDING:
 			CORBA_exception_set_system (
 				ev, ex_CORBA_TRANSIENT,
 				CORBA_COMPLETED_NO);
-			return;
+			break;
 
 		case PortableServer_POAManager_INACTIVE:
 			CORBA_exception_set_system (
 				ev, ex_CORBA_OBJ_ADAPTER,
 				CORBA_COMPLETED_NO);
-			return;
+			break;
 
 		case PortableServer_POAManager_ACTIVE:
 			break;
@@ -1015,6 +1149,11 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 			break;
 		}
 	}
+	if (ev->_major != CORBA_NO_EXCEPTION) {
+		return_exception (recv_buffer, m_data, ev);
+		goto clean_out;
+	}
+
 	oid = pobj->object_id;
 
 	if (!pobj->servant) {
@@ -1027,7 +1166,7 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 			break;
 
 		case PortableServer_USE_DEFAULT_SERVANT:
-			ORBit_POA_activate_object (
+			ORBit_POA_activate_object_T (
 				poa, pobj, poa->default_servant, ev);
 			break;
 
@@ -1042,8 +1181,10 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 	}
 
 	pobj->use_cnt++;
+	LINK_MUTEX_LOCK (poa->orb->lock);
 	poa->orb->current_invocations =
 		g_slist_prepend (poa->orb->current_invocations, pobj);
+	LINK_MUTEX_UNLOCK (poa->orb->lock);
 	
 	if (ev->_major == CORBA_NO_EXCEPTION && !pobj->servant)
 		CORBA_exception_set_system (
@@ -1088,6 +1229,9 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 		goto clean_out;
 	}
 
+	ORBit_RootObject_duplicate (poa);
+	POA_UNLOCK (poa);
+
 	if (recv_buffer) {
 		struct ORBit_POA_invoke_data invoke_data;
 
@@ -1100,6 +1244,9 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 	} else
 		small_skel (pobj->servant, ret, args, ctx, ev, imp);
 
+	POA_LOCK (poa);
+	ORBit_RootObject_release (poa);
+
  clean_out:
 	if (recv_buffer)
 		CORBA_exception_free (ev);
@@ -1119,26 +1266,30 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 			break;
 		}
 
-	g_assert ((ORBit_POAObject)poa->orb->current_invocations->data == pobj);
+	LINK_MUTEX_LOCK (poa->orb->lock);
+/*	g_assert ((ORBit_POAObject)poa->orb->current_invocations->data == pobj); */
 	poa->orb->current_invocations =
 		g_slist_remove (poa->orb->current_invocations, pobj);
+	LINK_MUTEX_UNLOCK (poa->orb->lock);
 	pobj->use_cnt--;
 
 	if (pobj->life_flags & ORBit_LifeF_NeedPostInvoke)
 		ORBit_POAObject_post_invoke (pobj);             
+
+	POA_UNLOCK (poa);
 }
 
 static ORBit_POAObject
-ORBit_POA_object_key_lookup (PortableServer_POA       poa,
-			     ORBit_ObjectKey         *objkey,
-			     PortableServer_ObjectId *object_id)
+ORBit_POA_object_key_lookup_T (PortableServer_POA       poa,
+			       ORBit_ObjectKey         *objkey,
+			       PortableServer_ObjectId *object_id)
 {
 	object_id->_buffer  = objkey->_buffer + ORBIT_ADAPTOR_PREFIX_LEN;
 	object_id->_length  = objkey->_length - ORBIT_ADAPTOR_PREFIX_LEN; 
 	object_id->_maximum = object_id->_length;
 	object_id->_release = CORBA_FALSE;
 
-	return ORBit_POA_object_id_lookup (poa, object_id);
+	return ORBit_POA_object_id_lookup_T (poa, object_id);
 }
 
 static void
@@ -1177,41 +1328,85 @@ ORBit_POAObject_invoke_incoming_request (ORBit_POAObject    pobj,
 	giop_recv_buffer_unuse (recv_buffer);
 }
 
+static inline void
+push_request_T (GIOPThread       *thread,
+		ORBit_POAObject  *pobj,
+		GIOPRecvBuffer  **recv_buffer)
+{
+	giop_thread_request_push
+		(thread, (gpointer *)pobj, (gpointer *)recv_buffer);
+}
+
+static void
+pool_push_request_for_T (gpointer          key_object,
+			 ORBit_POAObject  *pobj,
+			 GIOPRecvBuffer  **recv_buffer)
+{
+	g_assert (giop_thread_io ());
+	giop_thread_request_push_key
+		(key_object, (gpointer *)pobj, (gpointer *)recv_buffer);
+}			 
+
 typedef struct {
-	ORBit_POAObject pobj;
-	GIOPRecvBuffer *recv_buffer;
-} LocalClosure;
+	ORBit_POAObject  pobj;
+	GIOPRecvBuffer  *recv_buffer;
+} PoaIdleClosure;
 
 static gboolean
-local_main_handle (gpointer data)
+poa_invoke_at_idle (gpointer data)
 {
-	LocalClosure *lc = data;
+	PoaIdleClosure *pcl = data;
 
-	ORBit_POAObject_invoke_incoming_request (lc->pobj, lc->recv_buffer, NULL);
+	ORBit_POAObject_invoke_incoming_request
+		(pcl->pobj, pcl->recv_buffer, NULL);
 
-	g_free (lc);
+	g_free (data);
 
 	return FALSE;
 }
 
-/* Utter hack ... */
 static void
-queue_request (ORBit_POAObject *pobj,
-	       GIOPRecvBuffer **recv_buffer)
+push_request_idle (ORBit_POAObject  *pobj,
+		   GIOPRecvBuffer  **recv_buffer)
 {
-	LocalClosure *lc;
+	PoaIdleClosure *pcl = g_new (PoaIdleClosure, 1);
 
-	lc = g_new0 (LocalClosure, 1);
-
-	lc->pobj = *pobj;
-	lc->recv_buffer = *recv_buffer;
-
-	g_idle_add (local_main_handle, lc);
+	pcl->pobj = *pobj;
+	pcl->recv_buffer = *recv_buffer;
 
+	/* FIXME: track pending idle impls
+	 * for cleanup at shutdown ... */
+	g_idle_add (poa_invoke_at_idle, pcl);
+	
 	*pobj = NULL;
 	*recv_buffer = NULL;
 }
 
+static gboolean
+poa_recv_is_oneway (ORBit_POAObject pobj,
+		    GIOPRecvBuffer *recv_buffer)
+{
+	ORBit_IMethod            *m_data = NULL;
+	gpointer                  imp = NULL;
+	PortableServer_ClassInfo *klass;
+
+	g_return_val_if_fail (pobj != CORBA_OBJECT_NIL, FALSE);
+
+	klass = ORBIT_SERVANT_TO_CLASSINFO (pobj->servant);
+
+	if (!klass->impl_finder)
+		return FALSE;
+
+	klass->impl_finder ( pobj->servant,
+			     giop_recv_buffer_get_opname (recv_buffer),
+			     (gpointer *)&m_data, &imp);
+
+	if (m_data && m_data->flags & ORBit_I_METHOD_1_WAY)
+		return TRUE;
+
+	return FALSE;
+}
+
 static void
 ORBit_POA_handle_request (PortableServer_POA poa,
 			  GIOPRecvBuffer    *recv_buffer,
@@ -1223,7 +1418,9 @@ ORBit_POA_handle_request (PortableServer_POA poa,
 
 	CORBA_exception_init (&env);
 
-	pobj = ORBit_POA_object_key_lookup (poa, objkey, &object_id);
+	POA_LOCK (poa);
+
+	pobj = ORBit_POA_object_key_lookup_T (poa, objkey, &object_id);
 
 	if (!pobj)
 		switch (poa->p_request_processing) {
@@ -1236,7 +1433,7 @@ ORBit_POA_handle_request (PortableServer_POA poa,
 
 		case PortableServer_USE_DEFAULT_SERVANT: /* drop through */
 		case PortableServer_USE_SERVANT_MANAGER:
-			pobj = ORBit_POA_create_object (poa, &object_id, &env);
+			pobj = ORBit_POA_create_object_T (poa, &object_id, &env);
 			break;
 
 		default:
@@ -1251,40 +1448,63 @@ ORBit_POA_handle_request (PortableServer_POA poa,
 	else {
 		switch (poa->p_thread) {
 			case PortableServer_SINGLE_THREAD_MODEL:
-				if (giop_threaded ())
-					queue_request (&pobj, &recv_buffer);
+				if (giop_thread_io ())
+					push_request_T (giop_thread_get_main (),
+							&pobj, &recv_buffer);
 				break;
 			case PortableServer_ORB_CTRL_MODEL: {
 				ORBit_ObjectAdaptor adaptor = (ORBit_ObjectAdaptor) poa;
+				
+				switch (adaptor->thread_hint) {
+				case ORBIT_THREAD_HINT_PER_OBJECT:
+					pool_push_request_for_T (pobj, &pobj, &recv_buffer);
+					break;
 
-				switch (adaptor->thread_hint)
-				    {
-					case ORBIT_THREAD_HINT_ONEWAY_AT_IDLE:
-						/* FIXME: we need to lookup the
-						   method name, and cf. the oneway flag */
-					case ORBIT_THREAD_HINT_ALL_AT_IDLE:
-						queue_request (&pobj, &recv_buffer);
-						break;
-					case ORBIT_THREAD_HINT_NONE:
-						if (giop_threaded ())
-							queue_request (&pobj, &recv_buffer);
-						break;
-					default:
-						g_warning ("Binning incoming requests in threaded mode");
-						giop_recv_buffer_unuse (recv_buffer);
-						recv_buffer = NULL;
-						pobj = NULL;
-						break;
-				    }
-				break;
+				case ORBIT_THREAD_HINT_PER_POA:
+					pool_push_request_for_T (poa, &pobj, &recv_buffer);
+					break;
+
+				case ORBIT_THREAD_HINT_PER_CONNECTION:
+					pool_push_request_for_T (recv_buffer->connection,
+								 &pobj, &recv_buffer);
+					break;
+
+				case ORBIT_THREAD_HINT_PER_REQUEST:
+					pool_push_request_for_T (NULL, &pobj, &recv_buffer);
+					break;
+
+				case ORBIT_THREAD_HINT_ONEWAY_AT_IDLE:
+					if (!poa_recv_is_oneway (pobj, recv_buffer))
+						push_request_T (giop_thread_get_main (),
+								&pobj, &recv_buffer);
+					/* drop through */
+				case ORBIT_THREAD_HINT_ALL_AT_IDLE:
+					push_request_idle (&pobj, &recv_buffer);
+					break;
+
+				case ORBIT_THREAD_HINT_NONE:
+					if (giop_thread_io ())
+						push_request_T (giop_thread_get_main (),
+								&pobj, &recv_buffer);
+					break;
+				default:
+					g_warning ("Binning incoming requests in threaded mode");
+					giop_recv_buffer_unuse (recv_buffer);
+					recv_buffer = NULL;
+					pobj = NULL;
+					break;
 				}
-			default:
-				g_assert_not_reached ();
 				break;
 			}
+		default:
+			g_assert_not_reached ();
+			break;
+		}
 	}
-
+	
  send_sys_ex:
+	POA_UNLOCK (poa);
+
 	ORBit_POAObject_invoke_incoming_request (pobj, recv_buffer, &env);
 }
 
@@ -1431,7 +1651,7 @@ PortableServer_ObjectId_to_string (PortableServer_ObjectId *id,
 	poa_sys_exception_val_if_fail (memchr (id->_buffer, '\0', id->_length),
 				       ex_CORBA_BAD_PARAM, NULL);
 
-	str = CORBA_string_alloc (id->_length);
+	str = CORBA_string_alloc (id->_length + 1);
 	memcpy (str, id->_buffer, id->_length);
 	str [id->_length] = '\0';
 
@@ -1449,7 +1669,7 @@ PortableServer_ObjectId_to_wstring (PortableServer_ObjectId *id,
 	poa_sys_exception_val_if_fail (memchr (id->_buffer, '\0', id->_length),
 				       ex_CORBA_BAD_PARAM, NULL);
   
-	retval = CORBA_wstring_alloc (id->_length);
+	retval = CORBA_wstring_alloc (id->_length + 1);
 	for (i = 0; i < id->_length; i++)
 		retval [i] = id->_buffer [i];
 	retval [id->_length] = '\0';
@@ -1737,22 +1957,29 @@ PortableServer_POA_activate_object (PortableServer_POA            poa,
 				    const PortableServer_Servant  p_servant,
 				    CORBA_Environment            *ev)
 {
+	PortableServer_ObjectId    *result;
 	PortableServer_ServantBase *servant = p_servant;
 	ORBit_POAObject             newobj;
 
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, NULL);
 	poa_sys_exception_val_if_fail (servant != NULL, ex_CORBA_BAD_PARAM, NULL);
 
+	POA_LOCK (poa);
+
+	/* FIXME: unlock on preconditions */
 	poa_exception_val_if_fail (IS_RETAIN (poa), ex_PortableServer_POA_WrongPolicy, NULL);
 	poa_exception_val_if_fail (IS_SYSTEM_ID (poa), ex_PortableServer_POA_WrongPolicy, NULL);
 	poa_exception_val_if_fail (IS_MULTIPLE_ID (poa) || (IS_UNIQUE_ID (poa) && servant->_private == NULL),
 				   ex_PortableServer_POA_ServantAlreadyActive, NULL);
 
-	newobj = ORBit_POA_create_object (poa, NULL, ev);
-	ORBit_POA_activate_object (poa, newobj, servant, ev);
+	newobj = ORBit_POA_create_object_T (poa, NULL, ev);
+	ORBit_POA_activate_object_T (poa, newobj, servant, ev);
 
-	return (PortableServer_ObjectId *)
-			ORBit_sequence_CORBA_octet_dup (newobj->object_id);
+	result = ORBit_sequence_CORBA_octet_dup (newobj->object_id);
+
+	POA_UNLOCK (poa);
+
+	return result;
 }
 
 void
@@ -1768,9 +1995,11 @@ PortableServer_POA_activate_object_with_id (PortableServer_POA             poa,
 	poa_sys_exception_if_fail (objid != NULL, ex_CORBA_BAD_PARAM);
 	poa_sys_exception_if_fail (servant != NULL, ex_CORBA_BAD_PARAM);
 
+	POA_LOCK (poa);
+
 	poa_exception_if_fail (IS_RETAIN (poa), ex_PortableServer_POA_WrongPolicy);
 
-	pobj = ORBit_POA_object_id_lookup (poa, objid);
+	pobj = ORBit_POA_object_id_lookup_T (poa, objid);
 
 	if (pobj && pobj->servant)
 		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
@@ -1785,14 +2014,16 @@ PortableServer_POA_activate_object_with_id (PortableServer_POA             poa,
 		ORBit_POAObject newobj;
 
 		if (!pobj)
-			newobj = ORBit_POA_create_object (poa, objid, ev);
+			newobj = ORBit_POA_create_object_T (poa, objid, ev);
 		else
 			newobj = pobj;
 
-		ORBit_POA_activate_object (poa, newobj, servant, ev);
+		ORBit_POA_activate_object_T (poa, newobj, servant, ev);
 	}
 
 	ORBit_RootObject_release (pobj);
+
+	POA_UNLOCK (poa);
 }
 
 void
@@ -1805,13 +2036,17 @@ PortableServer_POA_deactivate_object (PortableServer_POA             poa,
 	poa_sys_exception_if_fail (poa != NULL, ex_CORBA_INV_OBJREF);
 	poa_sys_exception_if_fail (oid != NULL, ex_CORBA_BAD_PARAM);
 
+	POA_LOCK (poa);
+
 	poa_exception_if_fail (IS_RETAIN (poa), ex_PortableServer_POA_WrongPolicy);
 
-	pobj = ORBit_POA_object_id_lookup (poa, oid);
+	pobj = ORBit_POA_object_id_lookup_T (poa, oid);
 
 	if (pobj && pobj->servant)
 		ORBit_POA_deactivate_object (poa, pobj, CORBA_TRUE, CORBA_FALSE);
 
+	POA_UNLOCK (poa);
+
 	ORBit_RootObject_release (pobj);
 }
 
@@ -1820,15 +2055,22 @@ PortableServer_POA_create_reference (PortableServer_POA  poa,
 				     const CORBA_char   *intf,
 				     CORBA_Environment  *ev)
 {
+	CORBA_Object obj;
 	ORBit_POAObject pobj;
 
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, NULL);
 
+	POA_LOCK (poa);
+
 	poa_exception_val_if_fail (IS_SYSTEM_ID (poa), ex_PortableServer_POA_WrongPolicy, NULL);
 
-	pobj = ORBit_POA_create_object (poa, NULL, ev);
+	pobj = ORBit_POA_create_object_T (poa, NULL, ev);
 
-	return ORBit_POA_obj_to_ref (poa, pobj, intf, ev);
+	obj = ORBit_POA_obj_to_ref (poa, pobj, intf, ev);
+
+	POA_UNLOCK (poa);
+
+	return obj;
 }
 
 CORBA_Object
@@ -1843,9 +2085,11 @@ PortableServer_POA_create_reference_with_id (PortableServer_POA             poa,
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, NULL);
 	poa_sys_exception_val_if_fail (oid != NULL, ex_CORBA_BAD_PARAM, NULL);
 
-	pobj = ORBit_POA_object_id_lookup (poa, oid);
+	POA_LOCK (poa);
+
+	pobj = ORBit_POA_object_id_lookup_T (poa, oid);
 	if (!pobj)
-		newobj = ORBit_POA_create_object (poa, oid, ev);
+		newobj = ORBit_POA_create_object_T (poa, oid, ev);
 	else
 		newobj = CORBA_OBJECT_NIL;
 
@@ -1854,6 +2098,8 @@ PortableServer_POA_create_reference_with_id (PortableServer_POA             poa,
 	if (!newobj)
 		ORBit_RootObject_release (pobj);
 
+	POA_UNLOCK (poa);
+
 	return obj;
 }
 
@@ -1863,6 +2109,7 @@ PortableServer_POA_servant_to_id (PortableServer_POA            poa,
 				  CORBA_Environment            *ev)
 {
 	PortableServer_ServantBase *servant = p_servant;
+	PortableServer_ObjectId    *objid;
 	ORBit_POAObject             pobj = servant->_private;
 	gboolean                    defserv = IS_USE_DEFAULT_SERVANT (poa);
 	gboolean                    retain = IS_RETAIN (poa);
@@ -1872,19 +2119,23 @@ PortableServer_POA_servant_to_id (PortableServer_POA            poa,
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, NULL);
 	poa_sys_exception_val_if_fail (servant != NULL, ex_CORBA_BAD_PARAM, NULL);
 
+	POA_LOCK (poa);
+
 	poa_exception_val_if_fail (defserv || (retain && (unique || implicit)),
 				   ex_PortableServer_POA_WrongPolicy, NULL);
 
+	objid = NULL;
 	if (retain && unique && pobj && pobj->servant == servant)
-		return ORBit_sequence_CORBA_octet_dup (pobj->object_id);
+		objid = ORBit_sequence_CORBA_octet_dup (pobj->object_id);
 
 	else if (retain && implicit && (!unique || !pobj)) {
-		pobj = ORBit_POA_create_object (poa, NULL, ev);
-		ORBit_POA_activate_object (poa, pobj, servant, ev);
+		pobj = ORBit_POA_create_object_T (poa, NULL, ev);
+		ORBit_POA_activate_object_T (poa, pobj, servant, ev);
 
-		return ORBit_sequence_CORBA_octet_dup (pobj->object_id);
-	}
-	else {
+		objid = ORBit_sequence_CORBA_octet_dup (pobj->object_id);
+
+	} else {
+		GSList *l;
 		/*
 		 * FIXME:
 		 * This handles case 3 of the spec; but is broader:
@@ -1893,21 +2144,26 @@ PortableServer_POA_servant_to_id (PortableServer_POA            poa,
 		 * The stricter form could be implemented, 
 		 * but it would only add more code...
 		 */
-		GSList *l = poa->orb->current_invocations;
-
-		for ( ; l; l = l->next) {
+		LINK_MUTEX_LOCK (poa->orb->lock);
+		for (l = poa->orb->current_invocations; l; l = l->next) {
 			ORBit_POAObject pobj = l->data;
 
 			if (pobj->servant == servant)
-				return ORBit_sequence_CORBA_octet_dup (pobj->object_id);
+				objid = ORBit_sequence_CORBA_octet_dup (pobj->object_id);
 		}
+		LINK_MUTEX_UNLOCK (poa->orb->lock);
 	}
 
-	CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
-			     ex_PortableServer_POA_ServantNotActive,
-			     NULL);
+	if (!objid)
+		CORBA_exception_set
+			(ev, CORBA_USER_EXCEPTION,
+			 ex_PortableServer_POA_ServantNotActive,
+			 NULL);
 
-	return NULL;
+
+	POA_UNLOCK (poa);
+
+	return objid;
 }
 
 CORBA_Object
@@ -1917,6 +2173,7 @@ PortableServer_POA_servant_to_reference (PortableServer_POA            poa,
 {
 	PortableServer_ServantBase *servant = p_servant;
 	ORBit_POAObject             pobj = servant->_private;
+	CORBA_Object                result;
 	gboolean                    retain = IS_RETAIN (poa);
 	gboolean                    implicit = IS_IMPLICIT_ACTIVATION (poa);
 	gboolean                    unique = IS_UNIQUE_ID (poa);
@@ -1924,24 +2181,25 @@ PortableServer_POA_servant_to_reference (PortableServer_POA            poa,
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, CORBA_OBJECT_NIL);
 	poa_sys_exception_val_if_fail (servant != NULL, ex_CORBA_BAD_PARAM, CORBA_OBJECT_NIL);
 
+	POA_LOCK (poa);
 	poa_exception_val_if_fail (retain && (unique || implicit),
 				   ex_PortableServer_POA_WrongPolicy,
 				   CORBA_OBJECT_NIL);
 
+	result = CORBA_OBJECT_NIL;
 	if (retain && unique && pobj)
 		if (pobj->base.objref)
-			return ORBit_RootObject_duplicate (
-				pobj->base.objref);
+			result = ORBit_RootObject_duplicate (pobj->base.objref);
 		else
-			return ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
+			result = ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
 
 	else if (retain && implicit && (!unique || !pobj)) {
-		pobj = ORBit_POA_create_object (poa, NULL, ev);
-		ORBit_POA_activate_object (poa, pobj, servant, ev);
+		pobj = ORBit_POA_create_object_T (poa, NULL, ev);
+		ORBit_POA_activate_object_T (poa, pobj, servant, ev);
 
-		return ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
-	}
-	else {
+		result = ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
+	} else {
+		GSList *l;
 		/*
 		 * FIXME:
 		 * This case deals with "invoked in the context of
@@ -1951,21 +2209,25 @@ PortableServer_POA_servant_to_reference (PortableServer_POA            poa,
 		 * go backward from servant to pobj to use_cnt, but we
 		 * dont do this since forward search is more general 
 		 */
-		GSList *l = poa->orb->current_invocations;
-
-		for ( ; l; l = l->next) {
+		LINK_MUTEX_LOCK (poa->orb->lock);
+		for (l = poa->orb->current_invocations; l; l = l->next) {
 			ORBit_POAObject pobj = l->data;
 
 			if (pobj->servant == servant)
-				return ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
+				result = ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
 		}
+		LINK_MUTEX_UNLOCK (poa->orb->lock);
 	}
 
-	CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
-			     ex_PortableServer_POA_ServantNotActive,
-			     NULL);
+	if (result == CORBA_OBJECT_NIL)
+		CORBA_exception_set
+			(ev, CORBA_USER_EXCEPTION,
+			 ex_PortableServer_POA_ServantNotActive,
+			 NULL);
 
-	return NULL;
+	POA_UNLOCK (poa);
+
+	return result;
 }
 
 PortableServer_Servant
@@ -2027,30 +2289,34 @@ PortableServer_POA_id_to_servant (PortableServer_POA             poa,
 				  const PortableServer_ObjectId *object_id,
 				  CORBA_Environment             *ev)
 {
+	PortableServer_Servant servant = NULL;
+
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, NULL);
 	poa_sys_exception_val_if_fail (object_id != NULL, ex_CORBA_BAD_PARAM, NULL);
 
+	POA_LOCK (poa);
+
 	poa_exception_val_if_fail (IS_USE_DEFAULT_SERVANT (poa) || IS_RETAIN (poa),
 				   ex_PortableServer_POA_WrongPolicy, NULL);
 
 	if (IS_RETAIN (poa)) {
-		ORBit_POAObject pobj = ORBit_POA_object_id_lookup (poa, object_id);
-		PortableServer_Servant servant = NULL;
+		ORBit_POAObject pobj = ORBit_POA_object_id_lookup_T (poa, object_id);
 
 		if (pobj && pobj->servant)
 			servant = pobj->servant;
 
 		ORBit_RootObject_release (pobj);
 
-		if (servant)
-			return servant;
-
 	} else if (poa->default_servant)
-		return poa->default_servant;
+		servant = poa->default_servant;
 
-	CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
-			     ex_PortableServer_POA_ObjectNotActive,
-			     NULL);
+	if (!servant)
+		CORBA_exception_set
+			(ev, CORBA_USER_EXCEPTION,
+			 ex_PortableServer_POA_ObjectNotActive,
+			 NULL);
+
+	POA_UNLOCK (poa);
 
 	return NULL;
 }
@@ -2066,9 +2332,11 @@ PortableServer_POA_id_to_reference (PortableServer_POA             poa,
 	poa_sys_exception_val_if_fail (poa != NULL, ex_CORBA_INV_OBJREF, NULL);
 	poa_sys_exception_val_if_fail (object_id != NULL, ex_CORBA_BAD_PARAM, NULL);
 
+	POA_LOCK (poa);
+
 	poa_exception_val_if_fail (IS_RETAIN (poa), ex_PortableServer_POA_WrongPolicy, NULL);
 
-	pobj = ORBit_POA_object_id_lookup (poa, object_id);
+	pobj = ORBit_POA_object_id_lookup_T (poa, object_id);
 	if (!pobj || !pobj->servant) {
 		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
 				     ex_PortableServer_POA_ObjectNotActive, NULL);
@@ -2079,7 +2347,149 @@ PortableServer_POA_id_to_reference (PortableServer_POA             poa,
 		else
 			obj = ORBit_POA_obj_to_ref (poa, pobj, NULL, ev);
 	}
+
+	POA_UNLOCK (poa);
 	ORBit_RootObject_release (pobj);
 	
 	return obj;
 }
+
+void
+ORBit_poa_init (void)
+{
+	ORBit_class_assignment_lock = link_mutex_new ();
+	_ORBit_poa_manager_lock = link_mutex_new ();
+	giop_thread_set_main_handler (ORBit_POAObject_invoke_incoming_request);
+}
+
+gboolean
+ORBit_poa_allow_cross_thread_call (ORBit_POAObject pobj)
+{
+	gpointer key = NULL;
+	GIOPThread *self;
+	PortableServer_POA poa = pobj->poa;
+
+	if (!poa)
+		return TRUE;
+
+	self = giop_thread_self ();
+
+	switch (poa->p_thread) {
+	case PortableServer_SINGLE_THREAD_MODEL:
+		break;
+
+	case PortableServer_ORB_CTRL_MODEL: {
+		ORBit_ObjectAdaptor adaptor = (ORBit_ObjectAdaptor) poa;
+				
+		switch (adaptor->thread_hint) {
+		case ORBIT_THREAD_HINT_PER_OBJECT:
+			key = pobj;
+			break;
+			
+		case ORBIT_THREAD_HINT_PER_POA:
+			key = poa;
+			break;
+			
+		case ORBIT_THREAD_HINT_PER_CONNECTION:
+			/* FIXME: ? */
+		case ORBIT_THREAD_HINT_PER_REQUEST:
+			return TRUE;
+			break;
+			
+		case ORBIT_THREAD_HINT_ONEWAY_AT_IDLE:
+		case ORBIT_THREAD_HINT_ALL_AT_IDLE:
+		case ORBIT_THREAD_HINT_NONE:
+			break;
+		}
+		break;
+	}
+	}
+
+	giop_thread_new_check (self);
+
+	if (!key)
+		return (self == giop_thread_get_main ());
+	else
+		return giop_thread_same_key (key, TRUE);
+}
+
+static gpointer
+get_c_method (CORBA_Object                 obj,
+	      glong                        class_id,
+	      PortableServer_ServantBase **servant,
+	      glong                        method_offset)
+{
+	guchar *epv_start;
+	ORBit_POAObject pobj;
+
+	if (!obj ||
+	    !(pobj = (ORBit_POAObject) obj->adaptor_obj) ||
+	    !(*servant = (PortableServer_ServantBase *) pobj->servant))
+		return NULL;
+
+	if (method_offset <= 0 || class_id <= 0)
+		return NULL;
+
+	if (!ORBit_poa_allow_cross_thread_call (pobj))
+		return NULL;
+
+	if (ORBit_small_flags & ORBIT_SMALL_FORCE_GENERIC_MARSHAL)
+		return NULL;
+
+	/*
+	 * FIXME: we could propagate the size of the vepvmap_cache
+	 * to here, and check it to avoid some really bogus stuff.
+	 */
+	if (!class_id || !(pobj->base.interface->adaptor_type & ORBIT_ADAPTOR_POA) ||
+	    !pobj->vepvmap_cache)
+		return NULL;
+
+	epv_start = (guchar *) (*servant)->vepv [ pobj->vepvmap_cache [class_id] ];
+
+	if (!epv_start)
+		return NULL;
+
+	return *(gpointer *)(epv_start + method_offset);
+}
+
+void
+ORBit_c_stub_invoke (CORBA_Object        obj,
+		     ORBit_IMethods     *methods,
+		     glong               method_index,
+		     gpointer            ret,
+		     gpointer            args,
+		     CORBA_Context       ctx,
+		     CORBA_Environment  *ev,
+
+		     glong               class_id,
+		     glong               method_offset,
+		     ORBitSmallSkeleton  skel_impl)
+{
+	gpointer method_impl;
+	PortableServer_ServantBase *servant;
+
+	if (skel_impl &&
+	    (method_impl = get_c_method (obj, class_id,
+					 &servant, method_offset))) {
+		
+		/* Unwound PreCall
+		   ++( ((ORBit_POAObject)(obj)->adaptor_obj)->use_cnt );	    \
+		   (obj)->orb->current_invocations =                           \
+		   g_slist_prepend ((obj)->orb->current_invocations,   \
+		   (obj)->adaptor_obj);               \
+		*/
+
+		skel_impl (servant, ret, args, ctx, ev, method_impl);
+
+		/* Unwound PostCall
+		   (obj)->orb->current_invocations =                                                      \
+		   g_slist_remove ((obj)->orb->current_invocations, pobj);                        \
+		   --(((ORBit_POAObject)(obj)->adaptor_obj)->use_cnt);                                    \
+		   if (((ORBit_POAObject)(obj)->adaptor_obj)->life_flags & ORBit_LifeF_NeedPostInvoke)    \
+		   ORBit_POAObject_post_invoke (((ORBit_POAObject)(obj)->adaptor_obj));           \
+		*/
+	} else
+		ORBit_small_invoke_stub_n
+			(obj, methods, method_index,
+			 ret, args, ctx, ev);
+}
